using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;
using System.Collections;

namespace Sturm.CollectionWrapping {
	public abstract class TypedListWrapper : ITypedList {
		#region Creation
		#region Ctors
		public TypedListWrapper( ) {
		}

		public TypedListWrapper(PropertyDescriptorCollection propertyDescriptorCollection) {
			this.propertyDescriptorCollection = propertyDescriptorCollection;
		}
		#endregion

		#region Typed static creation helpers
		public static BindingList<T> CreateWrapper<T>(BindingList<T> wrappedList) {
			return CreateWrapper(wrappedList, null);
		}

		public static BindingList<T> CreateWrapper<T>(BindingList<T> wrappedList, PropertyDescriptorCollection propertyDescriptorCollection) {
			return new TypedListWrapperBindingListT<T>(wrappedList, propertyDescriptorCollection);
		}

		public static IBindingList CreateWrapper(IBindingList wrappedList) {
			return CreateWrapper(wrappedList, null);
		}

		public static IBindingList CreateWrapper(IBindingList wrappedList, PropertyDescriptorCollection propertyDescriptorCollection) {
			return new TypedListWrapperIBindingList(wrappedList, propertyDescriptorCollection);
		}

		public static IList<T> CreateWrapper<T>(IList<T> wrappedList) {
			return CreateWrapper<T>(wrappedList, null);
		}

		public static IList<T> CreateWrapper<T>(IList<T> wrappedList, PropertyDescriptorCollection propertyDescriptorCollection) {
			return new TypedListWrapperIListT<T>(wrappedList, propertyDescriptorCollection);
		}

		public static IList CreateWrapper(IList wrappedList) {
			return CreateWrapper(wrappedList, null);
		}

		public static IList CreateWrapper(IList wrappedList, PropertyDescriptorCollection propertyDescriptorCollection) {
			return new TypedListWrapperIList(wrappedList, propertyDescriptorCollection);
		}
		#endregion

		#region Untyped static creation helpers
		public static object CreateWrapper(object wrappedList) {
			return CreateWrapper(wrappedList, null);
		}

		public static object CreateWrapper(object wrappedList, PropertyDescriptorCollection propertyDescriptorCollection) {
			Type listType = wrappedList.GetType( );
			Type openWrapperType = null;
			Type[] wrapperTypeArguments = null;
			
			if (DerivesFromBindingListT(listType)) {
				openWrapperType = typeof(TypedListWrapperBindingListT<>);
				wrapperTypeArguments = GetBindingListTypeArguments(listType);
			}
			else if (ImplementsInterface(listType, typeof(IBindingList))) {
				openWrapperType = typeof(TypedListWrapperIBindingList);
			}
			else if (ImplementsInterface(listType, typeof(IList<>))) {
				openWrapperType = typeof(TypedListWrapperIListT<>);
				wrapperTypeArguments = GetIListTTypeArguments(listType);
			}
			else if (ImplementsInterface(listType, typeof(IList))) {
				openWrapperType = typeof(TypedListWrapperIList);
			}

			if (openWrapperType != null) {
				Type finalWrapperType = openWrapperType.ContainsGenericParameters ?
					openWrapperType.MakeGenericType(wrapperTypeArguments) :
					openWrapperType;
				
				return Activator.CreateInstance(finalWrapperType,
					new object[] { wrappedList, propertyDescriptorCollection });
			}

			return null;
		}

		#region Reflection helpers
		private static bool DerivesFromBindingListT(Type type) {
			if (type == typeof(object))
				return false;
			if (IsBindingListT(type))
				return true;
			return DerivesFromBindingListT(type.BaseType);
		}

		private static Type[] GetBindingListTypeArguments(Type type) {
			if (type == typeof(object))
				return null;
			if (IsBindingListT(type))
				return type.GetGenericArguments( );
			return GetBindingListTypeArguments(type.BaseType);
		}

		private static Type[] GetIListTTypeArguments(Type type) {
			Type[] interfaces = type.GetInterfaces( );
			Type ilistTType = typeof(IList<>);

			foreach (Type iface in interfaces) {
				if (iface.ContainsGenericParameters) {
					Type openIface = iface.GetGenericTypeDefinition( );
					if (openIface == ilistTType)
						return iface.GetGenericArguments( );
				}
			}
			return null;
		}

		private static bool IsOpenGenericType(Type type, Type openGenericType) {
			if (type.IsGenericType) {
				if (type.GetGenericTypeDefinition( ) == openGenericType)
					return true;
			}
			return false;
		}

		private static bool IsBindingListT(Type type) {
			return IsOpenGenericType(type, typeof(BindingList<>));
		}
		
		private static bool ImplementsInterface(Type type, Type interfaceType) {
			Type[] interfaces = type.GetInterfaces( );
			foreach (Type iface in interfaces) {
				if (iface == interfaceType ||
					(interfaceType.ContainsGenericParameters &&
					iface == interfaceType.GetGenericTypeDefinition( )))
					return true;
			}
			return false;
		}
		#endregion
		#endregion
		#endregion

		private PropertyDescriptorCollection propertyDescriptorCollection;
		public PropertyDescriptorCollection PropertyDescriptorCollection {
			get {
				return propertyDescriptorCollection;
			}
			set {
				propertyDescriptorCollection = value;
			}
		}

		private string listName = "Default";
		public string ListName {
			get {
				return listName;
			}
			set {
				listName = value;
			}
		}

		#region ITypedList Members

		public virtual PropertyDescriptorCollection GetItemProperties(PropertyDescriptor[] listAccessors) {
			return propertyDescriptorCollection;
		}

		public virtual string GetListName(PropertyDescriptor[] listAccessors) {
			return listName;
		}

		#endregion
	}
}